import cProfile
import datetime
import io
import os
import pathlib
import platform
import pstats
import re
import shutil
import subprocess
import sys
from abjad.system.AbjadObject import AbjadObject


class IOManager(AbjadObject):
    """
    Manages Abjad IO.
    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Managers'

    __slots__ = ()

    ### PRIVATE METHODS ###

    @staticmethod
    def _ensure_directory_existence(directory):
        if not directory:
            directory = '.'
        if not os.path.isdir(directory):
            lines = []
            line = 'Attention: {!r} does not exist on your system.'
            line = line.format(directory)
            lines.append(line)
            lines.append('Abjad will now create it to store all output files.')
            lines.append('Press any key to continue.')
            message = '\n'.join(lines)
            input(message)
            os.makedirs(directory)

    @staticmethod
    def _make_score_package(
        score_package_path,
        composer_email=None,
        composer_full_name=None,
        composer_last_name=None,
        composer_github_username=None,
        composer_library=None,
        score_title=None,
        year=None,
        ):
        import abjad
        configuration = abjad.AbjadConfiguration()
        score_package_name = os.path.basename(score_package_path)
        source_path = os.path.join(
            configuration.boilerplate_directory,
            'score',
            )
        target_path = score_package_path
        if not os.path.exists(target_path):
            shutil.copytree(source_path, target_path)
        else:
            for subentry in os.listdir(source_path):
                subentry_source = os.path.join(source_path, subentry)
                subentry_target = os.path.join(target_path, subentry)
                if os.path.isfile(subentry_source):
                    shutil.copy(subentry_source, subentry_target)
                elif os.path.isdir(subentry_source):
                    shutil.copytree(subentry_source, subentry_target)
                else:
                    raise ValueError(subentry_source)
        old_contents_directory = os.path.join(target_path, 'score')
        new_contents_directory = os.path.join(
            target_path,
            score_package_name,
            )
        shutil.move(old_contents_directory, new_contents_directory)
        suffixes = ('.py', '.tex', '.md', '.rst', '.ly', '.ily')
        for root, directory_name, file_names in os.walk(target_path):
            for file_name in file_names:
                if not file_name.endswith(suffixes):
                    continue
                file_ = os.path.join(root, file_name)
                with open(file_, 'r') as file_pointer:
                    template = file_pointer.read()
                try:
                    template = template.format(
                        composer_email=composer_email,
                        composer_full_name=composer_full_name,
                        composer_github_username=composer_github_username,
                        composer_last_name=composer_last_name,
                        composer_library=composer_library,
                        score_package_name=score_package_name,
                        score_title=score_title,
                        year=year,
                        )
                except (IndexError, KeyError):
                    lines = template.splitlines()
                    for i, line in enumerate(lines):
                        try:
                            lines[i] = line.format(
                                composer_email=composer_email,
                                composer_full_name=composer_full_name,
                                composer_github_username=composer_github_username,
                                composer_last_name=composer_last_name,
                                composer_library=composer_library,
                                score_package_name=score_package_name,
                                score_title=score_title,
                                year=year,
                                )
                        except (KeyError, IndexError, ValueError):
                            pass
                    template = '\n'.join(lines)
                with open(file_, 'w') as file_pointer:
                    file_pointer.write(template)

    @staticmethod
    def _read_from_pipe(pipe):
        lines = []
        string = pipe.read()
        for line in string.splitlines():
            try:
                line = line.decode('utf-8')
            except UnicodeDecodeError:
                continue
            lines.append(line)
        return '\n'.join(lines)

    @staticmethod
    def _warn_when_output_directory_almost_full(last_number):
        import abjad
        abjad_output_directory = abjad.abjad_configuration[
            'abjad_output_directory']
        max_number = 10000
        lines = []
        lines.append('')
        lines.append('WARNING: Abjad output directory almost full!')
        line = 'Abjad output directory contains {} files '
        line += 'and only {} are allowed.'
        line = line.format(last_number, max_number)
        lines.append(line)
        line = 'Please empty {} soon!'.format(abjad_output_directory)
        lines.append(line)
        lines.append('')
        for line in lines:
            print(line.center(80))

    ### PUBLIC METHODS ###

    @staticmethod
    def clear_terminal():
        """
        Clears terminal.

        Runs ``clear`` if OS is POSIX-compliant (UNIX / Linux / MacOS).

        Runs ``cls`` if OS is not POSIX-compliant (Windows).

        Returns none.
        """
        if os.name == 'posix':
            command = 'clear'
        else:
            command = 'cls'
        IOManager.spawn_subprocess(command)

    @staticmethod
    def count_function_calls(
        argument,
        global_context=None,
        local_context=None,
        fixed_point=True,
        ):
        """
        Counts function calls required to execute ``argument``.

        Wraps ``IOManager.profile(argument)``.

        Returns nonnegative integer.
        """
        def extract_count(profile_output):
            return int(profile_output.splitlines()[2].split()[0])
        if fixed_point:
            # profile at least twice to ensure consist results from profiler;
            # not sure why but profiler eventually levels off to consistent
            # output
            last_result, current_result = 'foo', 'bar'
            while current_result != last_result:
                last_result = current_result
                current_result = IOManager.profile(
                    argument,
                    print_to_terminal=False,
                    global_context=global_context,
                    local_context=local_context,
                    )
                current_result = extract_count(current_result)
            return current_result
        result = IOManager.profile(
            argument,
            print_to_terminal=False,
            global_context=global_context,
            local_context=local_context,
            )
        result = extract_count(result)
        return result

    @staticmethod
    def execute_file(path=None, attribute_names=None):
        """
        Executes file ``path``.

        Returns ``attribute_names`` from file.
        """
        assert path is not None
        assert isinstance(attribute_names, tuple)
        path = pathlib.Path(path)
        if not path.is_file():
            return
        file_contents_string = path.read_text()
        try:
            result = IOManager.execute_string(
                file_contents_string,
                attribute_names,
                )
        except:
            message = 'Exception raised in {}.'.format(path)
            # use print instead of display
            # to force to terminal even when called in silent context
            print(message)
            traceback.print_exc()
            return 'corrupt'
        return result

    @staticmethod
    def execute_string(
        string,
        attribute_names=None,
        local_namespace=None,
        ):
        """
        Executes ``string``.

        ..  container:: example

            >>> string = 'foo = 23'
            >>> attribute_names = ('foo', 'bar')
            >>> abjad.IOManager.execute_string(string, attribute_names)
            (23, None)

        Returns ``attribute_names`` from executed string.
        """
        assert isinstance(string, str)
        assert isinstance(attribute_names, tuple)
        if local_namespace is None:
            local_namespace = {}
        assert isinstance(local_namespace, dict)
        local_namespace = {}
        try:
            exec(string, local_namespace, local_namespace)
        except SyntaxError:
            return
        result = []
        for name in attribute_names:
            if name in local_namespace:
                result.append(local_namespace[name])
            else:
                result.append(None)
        result = tuple(result)
        return result

    @staticmethod
    def find_executable(name, flags=os.X_OK):
        """
        Finds executable ``name``.

        Similar to Unix ``which`` command.

        ..  container:: example

            >>> abjad.IOManager.find_executable('python3.6') # doctest: +SKIP
            ['/usr/bin/python3.6']

        Returns list of zero or more full paths to ``name``.
        """
        result = []
        extensions = [
            x
            for x in os.environ.get('PATHEXT', '').split(os.pathsep)
            if x
            ]
        path = os.environ.get('PATH', None)
        if path is None:
            return []
        for path in os.environ.get('PATH', '').split(os.pathsep):
            path = os.path.join(path, name)
            if os.access(path, flags):
                result.append(path)
            for extension in extensions:
                path_extension = path + extension
                if os.access(path_extension, flags):
                    result.append(path_extension)
        return result

    @staticmethod
    def get_last_output_file_name(
        extension=None,
        output_directory=None,
        ):
        """
        Gets last output file name in ``output_directory``.

        ..  container:: example

            >>> abjad.IOManager.get_last_output_file_name() # doctest: +SKIP
            '6222.ly'

        Gets last output file name in Abjad output directory when
        ``output_directory`` is none.

        Returns none when output directory contains no output files.

        Returns string or none.
        """
        import abjad
        pattern = re.compile('\d{4,4}.[a-z]{2,3}')
        output_directory = output_directory or abjad.abjad_configuration[
            'abjad_output_directory']
        if not os.path.exists(output_directory):
            return
        all_file_names = os.listdir(output_directory)
        if extension:
            all_output = [
                x for x in all_file_names
                if pattern.match(x) and x.endswith(extension)
                ]
        else:
            all_output = [x for x in all_file_names if pattern.match(x)]
        if all_output == []:
            last_output_file_name = None
        else:
            last_output_file_name = sorted(all_output)[-1]
        return last_output_file_name

    @staticmethod
    def get_next_output_file_name(
        file_extension='ly',
        output_directory=None,
        ):
        """
        Gets next output file name with ``file_extension`` in
        ``output_directory``.

        ..  container:: example

            >>> abjad.IOManager.get_next_output_file_name() # doctest: +SKIP
            '6223.ly'

        Gets next output file name with ``file_extension`` in Abjad output
        directory when ``output_directory`` is none.

        Returns string.
        """
        assert file_extension.isalpha() and \
            0 < len(file_extension) < 4, repr(file_extension)
        last_output = IOManager.get_last_output_file_name(
            output_directory=output_directory,
            )
        if last_output is None:
            next_number = 1
            next_output_file_name = '0001.{}'.format(file_extension)
        else:
            last_number = int(last_output.split('.')[0])
            next_number = last_number + 1
            next_output_file_name = '{next_number:04d}.{file_extension}'
            next_output_file_name = next_output_file_name.format(
                next_number=next_number,
                file_extension=file_extension,
                )
        if 9000 < next_number:
            IOManager._warn_when_output_directory_almost_full(last_number)
        return next_output_file_name

    @staticmethod
    def make_subprocess(command):
        """
        Makes Popen instance.

        ..  container:: example

            >>> command = 'echo "hellow world"'
            >>> abjad.IOManager.make_subprocess(command)
            <subprocess.Popen object at 0x...>

        Defined equal to:

            process = subprocess.Popen(
                command,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                )

        Redirects stderr to stdout.
        """
        return subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            )

    @staticmethod
    def open_file(file_path, application=None, line_number=None):
        """
        Opens ``file_path``.

        Uses ``application`` when ``application`` is not none.

        Uses Abjad configuration file ``text_editor`` when
        ``application`` is none.

        Takes best guess at operating system-specific file opener when both
        ``application`` and Abjad configuration file ``text_editor`` are none.

        Respects ``line_number`` when ``file_path`` can be opened with text
        editor.

        Returns none.
        """
        import abjad
        if sys.platform.lower().startswith('win'):
            os.startfile(file_path)
            return
        viewer = None
        if sys.platform.lower().startswith('linux'):
            viewer = application or 'xdg-open'
        elif file_path.endswith('.pdf'):
            viewer = application or abjad.abjad_configuration['pdf_viewer']
        elif file_path.endswith((
            '.log',
            '.py',
            '.rst',
            '.txt',
            )):
            viewer = application or abjad.abjad_configuration['text_editor']
        elif file_path.endswith((
            '.mid',
            '.midi',
            )):
            viewer = application or abjad.abjad_configuration['midi_player']
        viewer = viewer or 'open'
        if line_number:
            command = '{} +{} {}'.format(viewer, line_number, file_path)
        else:
            command = '{} {}'.format(viewer, file_path)
        IOManager.spawn_subprocess(command)

    @staticmethod
    def open_last_log():
        """
        Opens LilyPond log file in operating system-specific text editor.

        Returns none.
        """
        import abjad
        text_editor = abjad.abjad_configuration.get_text_editor()
        file_path = abjad.abjad_configuration.lilypond_log_file_path
        IOManager.open_file(file_path, application=text_editor)

    @staticmethod
    def open_last_ly(target=-1):
        """
        Opens last LilyPond output file produced by Abjad.

        Uses operating-specific text editor.

        Set ``target=-2`` to open the next-to-last LilyPond output file
        produced by Abjad, and so on.

        Returns none.
        """
        import abjad
        abjad_output_directory = abjad.abjad_configuration[
            'abjad_output_directory']
        text_editor = abjad.abjad_configuration.get_text_editor()
        if isinstance(target, int) and target < 0:
            last_lilypond = IOManager.get_last_output_file_name()
            if last_lilypond:
                last_number = last_lilypond
                last_number = last_number.replace('.ly', '')
                last_number = last_number.replace('.pdf', '')
                last_number = last_number.replace('.midi', '')
                last_number = last_number.replace('.mid', '')
                target_number = int(last_number) + (target + 1)
                target_str = '%04d' % target_number
                target_ly = os.path.join(
                    abjad_output_directory,
                    target_str + '.ly',
                    )
            else:
                print('Target LilyPond input file does not exist.')
        elif isinstance(target, int) and 0 <= target:
            target_str = '%04d' % target
            target_ly = os.path.join(
                abjad_output_directory,
                target_str + '.ly',
                )
        elif isinstance(target, str):
            target_ly = os.path.join(abjad_output_directory, target)
        else:
            message = 'can not get target LilyPond input from {}.'
            message = message.format(target)
            raise ValueError(message)
        if os.path.exists(target_ly):
            IOManager.open_file(target_ly, application=text_editor)
        else:
            message = 'Target LilyPond input file {} does not exist.'
            message = message.format(target_ly)
            print(message)

    @staticmethod
    def open_last_pdf(target=-1):
        """
        Opens last PDF generated by Abjad.

        Abjad writes PDFs to the ``~/.abjad/output`` directory by default.

        You may change this by setting the ``abjad_output_directory`` variable in
        the ``config.py`` file.

        Set ``target=-2`` to open the next-to-last PDF generated by Abjad.

        Returns none.
        """
        import abjad
        abjad_output_directory = abjad.abjad_configuration[
            'abjad_output_directory']
        if isinstance(target, int) and target < 0:
            last_lilypond_file_path = IOManager.get_last_output_file_name()
            if last_lilypond_file_path:
                result = os.path.splitext(last_lilypond_file_path)
                file_name_root, extension = result
                last_number = file_name_root
                target_number = int(last_number) + (target + 1)
                target_str = '%04d' % target_number
                target_pdf = os.path.join(
                    abjad_output_directory,
                    target_str + '.pdf',
                    )
            else:
                message = 'Target PDF does not exist.'
                print(message)
        elif isinstance(target, int) and 0 <= target:
            target_str = '%04d' % target
            target_pdf = os.path.join(
                abjad_output_directory,
                target_str + '.pdf',
                )
        elif isinstance(target, str):
            target_pdf = os.path.join(
                abjad_output_directory,
                target,
                )
        else:
            message = 'can not get target pdf name from {}.'
            message = message.format(target)
            raise ValueError(message)
        if os.stat(target_pdf):
            pdf_viewer = abjad.abjad_configuration['pdf_viewer']
            IOManager.open_file(target_pdf, pdf_viewer)
        else:
            message = 'target PDF {} does not exist.'
            message = message.format(target_pdf)
            print(message)

    @staticmethod
    def profile(
        argument,
        sort_by='cumulative',
        line_count=12, strip_dirs=True,
        print_callers=False,
        print_callees=False,
        global_context=None,
        local_context=None,
        print_to_terminal=True,
        ):
        """
        Profiles ``argument``.

        ..  container:: example

            ::

                >>> argument = 'abjad.Staff("c8 c8 c8 c8 c8 c8 c8 c8")'
                >>> abjad.IOManager.profile(
                ...     argument,
                ...     global_context=globals(),
                ...     ) # doctest: +SKIP
                Tue Apr  5 20:32:40 2011    _tmp_abj_profile

                        2852 function calls (2829 primitive calls) in 0.006 CPU seconds

                Ordered by: cumulative time
                List reduced from 118 to 12 due to restriction <12>

                ncalls  tottime  percall  cumtime  percall filename:lineno(function)
                        1    0.000    0.000    0.006    0.006 <string>:1(<module>)
                        1    0.001    0.001    0.003    0.003 make_notes.py:12(make_not
                        1    0.000    0.000    0.003    0.003 Staff.py:21(__init__)
                        1    0.000    0.000    0.003    0.003 Context.py:11(__init__)
                        1    0.000    0.000    0.003    0.003 Container.py:23(__init__)
                        1    0.000    0.000    0.003    0.003 Container.py:271(_initial
                        2    0.000    0.000    0.002    0.001 all_are_logical_voice_con
                    52    0.001    0.000    0.002    0.000 component_to_logical_voic
                        1    0.000    0.000    0.002    0.002 _construct_unprolated_not
                        8    0.000    0.000    0.002    0.000 make_tied_note.py:5(make_
                        8    0.000    0.000    0.002    0.000 make_tied_leaf.py:5(make_

        Wraps the built-in Python ``cProfile`` module.

        Set ``argument`` to any string of Abjad input.

        Set ``sort_by`` to ``'cumulative'``, ``'time'`` or ``'calls'``.

        Set ``line_count`` to any nonnegative integer.

        Set ``strip_dirs`` to true to strip directory names from output lines.

        See the `Python docs <http://docs.python.org/library/profile.html>`_
        for more information on the Python profilers.

        Returns none when ``print_to_terminal`` is false.

        Returns string when ``print_to_terminal`` is true.
        """
        now_string = datetime.datetime.today().strftime('%a %b %d %H:%M:%S %Y')
        profile = cProfile.Profile()
        if global_context is None:
            profile = profile.run(
                argument,
                )
        else:
            profile = profile.runctx(
                argument,
                global_context,
                local_context,
                )
        stats_stream = io.StringIO()
        stats = pstats.Stats(profile, stream=stats_stream)
        if sort_by == 'cum':
            sort_by = 'cumulative'
        if strip_dirs:
            stats.strip_dirs().sort_stats(sort_by).print_stats(line_count)
        else:
            stats.sort_stats(sort_by).print_stats(line_count)
        if print_callers:
            stats.sort_stats(sort_by).print_callers(line_count)
        if print_callees:
            stats.sort_stats(sort_by).print_callees(line_count)
        result = now_string + '\n\n' + stats_stream.getvalue()
        stats_stream.close()
        if print_to_terminal:
            print(result)
        else:
            return result

    @staticmethod
    def run_command(command):
        """
        Makes subprocess with ``command``.

        Runs command.

        Displays nothing.

        Returns stdout from subprocess.
        """
        process = IOManager.make_subprocess(command)
        lines = IOManager._read_from_pipe(process.stdout)
        lines = lines.splitlines()
        return lines

    @staticmethod
    def run_lilypond(ly_path, flags=None, lilypond_path=None):
        """
        Runs LilyPond on ``ly_path``.

        Writes redirected output of Unix ``date`` to top line of LilyPond log
        file.

        Then appends redirected output of LilyPond output to the LilyPond log
        file.

        Returns none.
        """
        import abjad
        ly_path = str(ly_path)
        lilypond_path = abjad.abjad_configuration.get('lilypond_path')
        if not lilypond_path:
            lilypond_path = abjad.IOManager.find_executable('lilypond')
            if lilypond_path:
                lilypond_path = lilypond_path[0]
            else:
                lilypond_path = 'lilypond'
        lilypond_base, extension = os.path.splitext(ly_path)
        flags = flags or ''
        date = datetime.datetime.now().strftime('%c')
        log_file_path = abjad.abjad_configuration.lilypond_log_file_path
        command = '{} {} -dno-point-and-click -o {} {}'.format(
            lilypond_path,
            flags,
            lilypond_base,
            ly_path,
            )
        process = subprocess.Popen(
            command,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            )
        subprocess_output, _ = process.communicate()
        subprocess_output = subprocess_output.decode('utf-8')
        exit_code = process.returncode
        with open(log_file_path, 'w') as file_pointer:
            file_pointer.write(date + '\n')
            file_pointer.write(subprocess_output)
        postscript_path = ly_path.replace('.ly', '.ps')
        try:
            os.remove(postscript_path)
        except OSError:
            pass
        # TODO: maybe just 'return exit_code'?
        if exit_code:
            return False
        return True

    @staticmethod
    def save_last_ly_as(file_path):
        """
        Saves last LilyPond file created by Abjad as ``file_path``.

        Returns none.
        """
        import abjad
        abjad_output_directory = abjad.abjad_configuration[
            'abjad_output_directory']
        last_output_file_name = IOManager.get_last_output_file_name(
            extension='.ly',
            )
        if last_output_file_name is None:
            return
        last_ly_full_name = os.path.join(
            abjad_output_directory,
            last_output_file_name,
            )
        with open(file_path, 'w') as new:
            with open(last_ly_full_name, 'r') as old:
                new.write(''.join(old.readlines()))

    @staticmethod
    def save_last_pdf_as(file_path):
        """
        Saves last PDF created by Abjad as ``file_path``.

        Returns none.
        """
        import abjad
        abjad_output_directory = abjad.abjad_configuration[
            'abjad_output_directory']
        last_output_file_name = IOManager.get_last_output_file_name(
            extension='.pdf',
            )
        last_pdf_full_name = os.path.join(
            abjad_output_directory,
            last_output_file_name,
            )
        with open(file_path, 'w') as new:
            with open(last_pdf_full_name, 'r') as old:
                new.write(''.join(old.readlines()))

    @staticmethod
    def spawn_subprocess(command):
        """
        Spawns subprocess and runs ``command``.

        The function is basically a reimplementation of the
        deprecated ``os.system()`` using Python's ``subprocess`` module.

        Redirects stderr to stdout.

        Returns integer exit code.
        """
        return subprocess.call(command, shell=True)



import abc
import copy
import uqbar.graphs
from abjad import enums
from abjad import exceptions
from abjad.indicators.MetronomeMark import MetronomeMark
from abjad.mathtools.NonreducedFraction import NonreducedFraction
from abjad.mathtools.Ratio import Ratio
from abjad.system.FormatSpecification import FormatSpecification
from abjad.system.LilyPondFormatManager import LilyPondFormatManager
from abjad.top.attach import attach
from abjad.top.detach import detach
from abjad.top.inspect import inspect
from abjad.top.override import override
from abjad.top.select import select
from abjad.top.sequence import sequence
from abjad.top.setting import setting
from abjad.utilities.Duration import Duration
from abjad.utilities.Multiplier import Multiplier
from .Component import Component


class Leaf(Component):
    """
    Leaf baseclass.

    Leaves include notes, rests, chords and skips.
    """

    ### CLASS VARIABLES ##

    __slots__ = (
        '_after_grace_container',
        '_grace_container',
        '_leaf_index',
        '_spanners',
        '_written_duration',
        )

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, written_duration):
        Component.__init__(self)
        self._after_grace_container = None
        self._grace_container = None
        self._leaf_index = None
        self._spanners = []
        self.written_duration = Duration(written_duration)

    ### SPECIAL METHODS ###

    def __copy__(self, *arguments):
        """
        Shallow copies leaf.

        Returns new leaf.
        """
        new = Component.__copy__(self, *arguments)
        grace_container = self._grace_container
        if grace_container is not None:
            new_grace_container = grace_container._copy_with_children()
            attach(new_grace_container, new)
        after_grace_container = self._after_grace_container
        if after_grace_container is not None:
            new_after_grace_container = \
                after_grace_container._copy_with_children()
            attach(new_after_grace_container, new)
        return new

    def __getnewargs__(self):
        """
        Gets new arguments.

        Returns tuple.
        """
        return (self.written_duration,)

    def __str__(self):
        """
        Gets string representation of leaf.

        Returns string.
        """
        return self._get_compact_representation()

    ### PRIVATE METHODS ###

    def _append_spanner(self, spanner):
        if id(spanner) in [id(_) for _ in self._spanners]:
            return
        self._spanners.append(spanner)

    def _as_graphviz_node(self):
        lilypond_format = self._get_compact_representation()
        lilypond_format = lilypond_format.replace('<', '&lt;')
        lilypond_format = lilypond_format.replace('>', '&gt;')
        node = Component._as_graphviz_node(self)
        node[0].extend([
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    type(self).__name__,
                    attributes={'border': 0},
                    ),
                ]),
            uqbar.graphs.HRule(),
            uqbar.graphs.TableRow([
                uqbar.graphs.TableCell(
                    lilypond_format,
                    attributes={'border': 0},
                    ),
                ]),
            ])
        return node

    def _copy_override_and_set_from_leaf(self, leaf):
        if getattr(leaf, '_overrides', None) is not None:
            self._overrides = copy.copy(override(leaf))
        if getattr(leaf, '_lilypond_setting_name_manager', None) is not None:
            self._lilypond_setting_name_manager = copy.copy(
                setting(leaf))
        new_wrappers = []
        for wrapper in leaf._wrappers:
            new_wrapper = copy.copy(wrapper)
            new_wrappers.append(new_wrapper)
        for new_wrapper in new_wrappers:
            attach(new_wrapper, self)

    def _detach_after_grace_container(self):
        if self._after_grace_container is not None:
            return detach(self._after_grace_container, self)

    def _detach_grace_container(self):
        if self._grace_container is not None:
            return detach(self._grace_container, self)

    def _detach_spanners(self, prototype=None):
        spanners = self._get_spanners(prototype=prototype)
        for spanner in spanners:
            spanner._sever_all_leaves()
        return spanners

    def _format_after_grace_body(self):
        result = []
        if self._after_grace_container is not None:
            after_grace = self._after_grace_container
            if len(after_grace):
                result.append(format(after_grace))
        return ['after grace body', result]

    def _format_after_grace_opening(self):
        result = []
        if self._after_grace_container is not None:
            if len(self._after_grace_container):
                result.append(r'\afterGrace')
        return ['after grace opening', result]

    def _format_absolute_after_slot(self, bundle):
        result = []
        result.append(('literals', bundle.absolute_after.commands))
        return result

    def _format_absolute_before_slot(self, bundle):
        result = []
        result.append(('literals', bundle.absolute_before.commands))
        return result

    def _format_after_slot(self, bundle):
        result = []
        result.append(('stem_tremolos', bundle.after.stem_tremolos))
        result.append(('articulations', bundle.after.articulations))
        result.append(('markup', bundle.after.markup))
        result.append(('spanners', bundle.after.spanners))
        result.append(('spanner_stops', bundle.after.spanner_stops))
        result.append(('spanner_starts', bundle.after.spanner_starts))
        # NOTE: LilyPond demands \startTrillSpan appear after almost all
        #       other format contributions; pitched trills dangerously
        #       suppress markup and the starts of other spanners when
        #       \startTrillSpan appears lexically prior to those commands;
        #       but \startTrillSpan must appear before calls to \set.
        result.append(('trill_spanner_starts', bundle.after.trill_spanner_starts))
        result.append(('commands', bundle.after.commands))
        result.append(('commands', bundle.after.leaks))
        result.append(('comments', bundle.after.comments))
        return result

    def _format_before_slot(self, bundle):
        result = []
        result.append(self._format_grace_body())
        result.append(('comments', bundle.before.comments))
        result.append(('commands', bundle.before.commands))
        result.append(('indicators', bundle.before.indicators))
        result.append(('grob reverts', bundle.grob_reverts))
        result.append(('grob overrides', bundle.grob_overrides))
        result.append(('context settings', bundle.context_settings))
        result.append(('spanners', bundle.before.spanners))
        return result

    def _format_close_brackets_slot(self, bundle):
        return []

    def _format_closing_slot(self, bundle):
        result = []
        result.append(self._format_after_grace_body())
        result.append(('spanners', bundle.closing.spanners))
        result.append(('commands', bundle.closing.commands))
        result.append(('indicators', bundle.closing.indicators))
        result.append(('comments', bundle.closing.comments))
        return result

    def _format_contents_slot(self, bundle):
        result = []
        result.append(self._format_leaf_body(bundle))
        return result

    def _format_grace_body(self):
        result = []
        if self._grace_container is not None:
            grace = self._grace_container
            if len(grace):
                result.append(format(grace))
        return ['grace body', result]

    def _format_leaf_body(self, bundle):
        indent = LilyPondFormatManager.indent
        result = self._format_leaf_nucleus()[1]
        return ['self body', result]

    def _format_leaf_nucleus(self):
        return ['nucleus', self._get_body()]

    def _format_open_brackets_slot(self, bundle):
        return []

    def _format_opening_slot(self, bundle):
        result = []
        result.append(('comments', bundle.opening.comments))
        result.append(('indicators', bundle.opening.indicators))
        result.append(('commands', bundle.opening.commands))
        result.append(self._format_after_grace_opening())
        result.append(('spanners', bundle.opening.spanners))
        return result

    def _get_compact_representation(self):
        return f'({self._get_formatted_duration()})'

    def _get_format_pieces(self):
        return self._get_lilypond_format().split('\n')

    def _get_format_specification(self):
        summary = self._get_compact_representation()
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            repr_args_values=[summary],
            storage_format_args_values=[format(self, 'lilypond')],
            storage_format_is_indented=False,
            storage_format_kwargs_names=[],
            )

    def _get_leaf(self, n):

        def next(component):
            new_component = component._get_nth_component_in_time_order_from(1)
            if new_component is None:
                return
            candidates = new_component._get_descendants_starting_with()
            candidates = [
                x for x in candidates if isinstance(x, Leaf)
                ]
            for candidate in candidates:
                selection = select([component, candidate])
                if selection.are_logical_voice():
                    return candidate

        def previous(component):
            new_component = component._get_nth_component_in_time_order_from(-1)
            if new_component is None:
                return
            candidates = new_component._get_descendants_stopping_with()
            candidates = [
                x for x in candidates if isinstance(x, Leaf)
                ]
            for candidate in candidates:
                selection = select([component, candidate])
                if selection.are_logical_voice():
                    return candidate
        current_leaf = self
        if n < 0:
            for i in range(abs(n)):
                current_leaf = previous(current_leaf)
                if current_leaf is None:
                    break
        elif n == 0:
            pass
        else:
            for i in range(n):
                current_leaf = next(current_leaf)
                if current_leaf is None:
                    break
        return current_leaf

    def _get_logical_tie(self):
        import abjad
        for component in [self]:
            ties = inspect(component).spanners(abjad.Tie)
            if len(ties) == 1:
                tie = ties.pop()
                return abjad.LogicalTie(items=tie.leaves)
            elif 1 < len(ties):
                message = f'parentage of {self!r} contains {len(ties)} ties.'
                raise Exception(ties)
        else:
            return abjad.LogicalTie(items=self)

    def _get_spanner(self, prototype=None):
        spanners = self._get_spanners(prototype=prototype)
        if not spanners:
            raise exceptions.MissingSpannerError('no spanner found.')
        elif len(spanners) == 1:
            return spanners.pop()
        else:
            message = f'multiple spanners found: {spanners!r}'
            raise ExtraSpannerError(message)

    def _get_spanners(self, prototype=None):
        import abjad
        prototype = prototype or (abjad.Spanner,)
        if not isinstance(prototype, tuple):
            prototype = (prototype, )
        spanner_items = prototype[:]
        prototype, spanner_objects = [], []
        for spanner_item in spanner_items:
            if isinstance(spanner_item, type):
                prototype.append(spanner_item)
            elif isinstance(spanner_item, abjad.Spanner):
                spanner_objects.append(spanner_item)
            else:
                message = 'must be spanner class or spanner object: {!r}'
                message = message.format(spanner_item)
        prototype = tuple(prototype)
        spanner_objects = tuple(spanner_objects)
        matching_spanners = []
        for spanner in self._spanners:
            if isinstance(spanner, prototype):
                if id(spanner) not in [id(_) for _ in matching_spanners]:
                    matching_spanners.append(spanner)
            elif any(spanner == x for x in spanner_objects):
                if id(spanner) not in [id(_) for _ in matching_spanners]:
                    matching_spanners.append(spanner)
        return matching_spanners

    def _has_spanner(self, prototype=None):
        spanners = self._get_spanners(prototype=prototype)
        return bool(spanners)

    def _process_contribution_packet(self, contribution_packet):
        manager = LilyPondFormatManager
        indent = manager.indent
        result = ''
        for contributor, contributions in contribution_packet:
            if contributions:
                if isinstance(contributor, tuple):
                    contributor = indent + contributor[0] + ':\n'
                else:
                    contributor = indent + contributor + ':\n'
                result += contributor
                for contribution in contributions:
                    contribution = (indent * 2) + contribution + '\n'
                    result += contribution
        return result

    def _remove_spanner(self, spanner):
        if id(spanner) not in [id(_) for _ in self._spanners]:
            raise Exception(f'{self!s} has no {spanner}.')
        spanners = [_ for _ in self._spanners if id(_) != id(spanner)]
        self._spanners = spanners

    def _report_format_contributions(self):
        manager = LilyPondFormatManager
        indent = manager.indent
        bundle = manager.bundle_format_contributions(self)
        report = ''
        report += 'slot absolute before:\n'
        packet = self._format_absolute_before_slot(bundle)
        report += self._process_contribution_packet(packet)
        report += 'slot 1:\n'
        packet = self._format_before_slot(bundle)
        report += self._process_contribution_packet(packet)
        report += 'slot 3:\n'
        packet = self._format_opening_slot(bundle)
        report += self._process_contribution_packet(packet)
        report += 'slot 4:\n'
        report += indent + 'leaf body:\n'
        string = self._format_contents_slot(bundle)[0][1][0]
        report += (2 * indent) + string + '\n'
        report += 'slot 5:\n'
        packet = self._format_closing_slot(bundle)
        report += self._process_contribution_packet(packet)
        report += 'slot 7:\n'
        packet = self._format_after_slot(bundle)
        report += self._process_contribution_packet(packet)
        report += 'slot absolute after:\n'
        packet = self._format_absolute_after_slot(bundle)
        report += self._process_contribution_packet(packet)
        while report[-1] == '\n':
            report = report[:-1]
        return report

    def _scale(self, multiplier):
        new_duration = multiplier * self._get_duration()
        self._set_duration(new_duration)

    def _set_duration(self, new_duration, repeat_ties=False):
        import abjad
        new_duration = Duration(new_duration)
        # change LilyPond multiplier if leaf already has LilyPond multiplier
        if self._get_indicators(Multiplier):
            detach(Multiplier, self)
            multiplier = new_duration.__div__(self.written_duration)
            attach(multiplier, self)
            return select(self)
        # change written duration if new duration is assignable
        try:
            self.written_duration = new_duration
            return select(self)
        except exceptions.AssignabilityError:
            pass
        # make new notes or tuplets if new duration is nonassignable
        maker = abjad.NoteMaker(
            repeat_ties=repeat_ties,
            )
        components = maker(0, new_duration)
        if isinstance(components[0], Leaf):
            tied_leaf_count = len(components) - 1
            tied_leaves = tied_leaf_count * self
            all_leaves = [self] + tied_leaves
            for leaf, component in zip(all_leaves, components):
                leaf.written_duration = component.written_duration
            self._splice(tied_leaves, grow_spanners=True)
            if not inspect(self).has_spanner(abjad.Tie):
                tie = abjad.Tie()
                if tie._attachment_test(self):
                    tie = abjad.Tie(repeat=repeat_ties)
                    attach(tie, all_leaves)
            return select(all_leaves)
        else:
            assert isinstance(components[0], abjad.Tuplet)
            tuplet = components[0]
            components = tuplet[:]
            tied_leaf_count = len(components) - 1
            tied_leaves = tied_leaf_count * self
            all_leaves = [self] + tied_leaves
            for leaf, component in zip(all_leaves, components):
                leaf.written_duration = component.written_duration
            self._splice(tied_leaves, grow_spanners=True)
            if not inspect(self).has_spanner(abjad.Tie):
                tie = abjad.Tie()
                if tie._attachment_test(self):
                    tie = abjad.Tie(repeat=repeat_ties)
                    attach(tie, all_leaves)
            multiplier = tuplet.multiplier
            tuplet = abjad.Tuplet(multiplier, [])
            abjad.mutate(all_leaves).wrap(tuplet)
            return select(tuplet)

    def _split_by_durations(
        self,
        durations,
        cyclic=False,
        fracture_spanners=False,
        tie_split_notes=True,
        repeat_ties=False,
        ):
        import abjad
        durations = [Duration(_) for _ in durations]
        durations = sequence(durations)
        leaf_duration = inspect(self).duration()
        if cyclic:
            durations = durations.repeat_to_weight(leaf_duration)
        if sum(durations) < leaf_duration:
            last_duration = leaf_duration - sum(durations)
            durations = list(durations)
            durations.append(last_duration)
            durations = sequence(durations)
        durations = durations.truncate(weight=leaf_duration)
        result_selections = []
        # detach grace containers
        grace_container = self._detach_grace_container()
        after_grace_container = self._detach_after_grace_container()
        leaf_prolation = inspect(self).parentage().prolation
        for duration in durations:
            new_leaf = copy.copy(self)
            preprolated_duration = duration / leaf_prolation
            selection = new_leaf._set_duration(
                preprolated_duration,
                repeat_ties=repeat_ties,
                )
            result_selections.append(selection)
        result_components = sequence(result_selections).flatten(depth=-1)
        result_components = select(result_components)
        result_leaves = select(result_components).leaves()
        assert all(isinstance(_, abjad.Selection) for _ in result_selections)
        assert all(isinstance(_, Component) for _ in result_components)
        assert result_leaves.are_leaves()
        if inspect(self).has_spanner(abjad.Tie):
            for leaf in result_leaves:
                detach(abjad.Tie, leaf)
        # strip result leaves of indicators (other than multipliers)
        for leaf in result_leaves:
            multiplier = inspect(leaf).indicator(Multiplier)
            detach(object, leaf)
            if multiplier is not None:
                attach(multiplier, leaf)
        # replace leaf with flattened result
        selection = select(self)
        parent, start, stop = selection._get_parent_and_start_stop_indices()
        if parent:
            parent.__setitem__(slice(start, stop + 1), result_components)
        else:
            selection._give_dominant_spanners(result_components)
            selection._withdraw_from_crossing_spanners()
        # fracture spanners
        if fracture_spanners:
            first_selection = result_selections[0]
            for spanner in inspect(first_selection[-1]).spanners():
                index = spanner._index(first_selection[-1])
                spanner._fracture(index, direction=enums.Right)
            last_selection = result_selections[-1]
            for spanner in inspect(last_selection[0]).spanners():
                index = spanner._index(last_selection[0])
                spanner._fracture(index, direction=enums.Left)
            for middle_selection in result_selections[1:-1]:
                spanners = inspect(middle_selection[0]).spanners()
                for spanner in spanners:
                    index = spanner._index(middle_selection[0])
                    spanner._fracture(index, direction=enums.Left)
                spanners = inspect(middle_selection[-1]).spanners()
                for spanner in spanners:
                    index = spanner._index(middle_selection[-1])
                    spanner._fracture(index, direction=enums.Right)
        # move indicators
        first_result_leaf = result_leaves[0]
        last_result_leaf = result_leaves[-1]
        for indicator in inspect(self).indicators():
            if isinstance(indicator, Multiplier):
                continue
            detach(indicator, self)
            direction = getattr(indicator, '_time_orientation', enums.Left)
            if direction is enums.Left:
                attach(indicator, first_result_leaf)
            elif direction == enums.Right:
                attach(indicator, last_result_leaf)
            else:
                raise ValueError(direction)
        # move grace containers
        if grace_container is not None:
            container = grace_container[0]
            assert isinstance(container, abjad.GraceContainer), repr(container)
            attach(container, first_result_leaf)
        if after_grace_container is not None:
            container = after_grace_container[0]
            prototype = abjad.AfterGraceContainer
            assert isinstance(container, prototype), repr(container)
            attach(container, last_result_leaf)
        if isinstance(result_components[0], abjad.Tuplet):
            abjad.mutate(result_components).fuse()
        # tie split notes
        if isinstance(self, (abjad.Note, abjad.Chord)) and tie_split_notes:
            result_leaves._attach_tie_to_leaves(
                repeat_ties=repeat_ties,
                )
        assert isinstance(result_selections, list), repr(result_selections)
        assert all(isinstance(_, abjad.Selection) for _ in result_selections)
        return result_selections

    ### PRIVATE PROPERTIES ###

    def _get_duration_in_seconds(self):
        mark = self._get_effective(MetronomeMark)
        if mark is not None and not mark.is_imprecise:
            result = (
                self._get_duration() /
                mark.reference_duration /
                mark.units_per_minute * 60
                )
            return Duration(result)
        raise exceptions.MissingMetronomeMarkError

    def _get_formatted_duration(self):
        duration_string = self.written_duration.lilypond_duration_string
        multiplier = None
        multiplier_prototype = (Multiplier, NonreducedFraction)
        multipliers = self._get_indicators(multiplier_prototype)
        if not multipliers:
            pass
        elif len(multipliers) == 1:
            multiplier = multipliers[0]
        elif 1 < len(multipliers):
            message = 'more than one LilyPond duration multiplier.'
            raise ValueError(message)
        if multiplier is not None:
            result = '{} * {!s}'.format(duration_string, multiplier)
        else:
            result = duration_string
        return result

    def _get_multiplied_duration(self):
        if self.written_duration:
            multiplier_prototype = (Multiplier, NonreducedFraction)
            if self._get_indicators(multiplier_prototype):
                multipliers = self._get_indicators(multiplier_prototype)
                if 1 == len(multipliers):
                    multiplier = multipliers[0]
                    multiplier = Duration(multiplier)
                elif 1 < len(multipliers):
                    message = 'more than one duration multiplier.'
                    raise ValueError(message)
                multiplied_duration = multiplier * self.written_duration
                return multiplied_duration
            else:
                return Duration(self.written_duration)
        else:
            return None

    def _get_preprolated_duration(self):
        return self._get_multiplied_duration()

    ### PUBLIC PROPERTIES ###

    @property
    def written_duration(self):
        """
        Written duration of leaf.

        Set to duration.

        Returns duration.
        """
        return self._written_duration

    @written_duration.setter
    def written_duration(self, argument):
        rational = Duration(argument)
        if not rational.is_assignable:
            message = 'not assignable duration: {!r}.'
            message = message.format(rational)
            raise exceptions.AssignabilityError(message)
        self._written_duration = rational



def play(argument):
    r"""
    Plays ``argument``.

    ..  container:: example

        >>> note = abjad.Note("c'4")
        >>> abjad.play(note) # doctest: +SKIP

    Makes MIDI file.

    Appends ``.mid`` filename extension under Windows.

    Appends ``.midi`` filename extension under other operating systems.

    Opens MIDI file.

    Returns none.
    """
    import abjad
    from abjad import abjad_configuration
    assert hasattr(argument, '__illustrate__')
    result = abjad.persist(argument).as_midi()
    midi_file_path, abjad_formatting_time, lilypond_rendering_time = result
    midi_player = abjad_configuration['midi_player']
    abjad.IOManager.open_file(midi_file_path, midi_player)



def annotate(component, annotation, indicator):
    r"""
    Annotates ``component`` with ``indicator``.

    ..  container:: example

        Annotates first note in staff:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.annotate(staff[0], 'bow_direction', abjad.Down)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                c'4
                d'4
                e'4
                f'4
            }

        >>> abjad.inspect(staff[0]).annotation('bow_direction')
        Down

        >>> abjad.inspect(staff[0]).annotation('bow_fraction') is None
        True

        >>> abjad.inspect(staff[0]).annotation('bow_fraction', 99)
        99

    Returns none.
    """
    import abjad
    assert isinstance(annotation, str), repr(annotation)
    wrapper = abjad.Wrapper(
        annotation=annotation,
        component=component,
        indicator=indicator,
        )
    wrapper._bind_to_component(component)



try:
    import quicktions as fractions # type: ignore
except ImportError:
    import fractions # type: ignore
import typing
from abjad import system
from abjad.system.AbjadObject import AbjadObject


class NonreducedFraction(AbjadObject, fractions.Fraction):
    """
    Nonreduced fraction.

    ..  container:: example

        Initializes with an integer numerator and integer denominator:

        >>> abjad.NonreducedFraction(3, 6)
        NonreducedFraction(3, 6)

    ..  container:: example

        Initializes with only an integer denominator:

        >>> abjad.NonreducedFraction(3)
        NonreducedFraction(3, 1)

    ..  container:: example

        Initializes with an integer pair:

        >>> abjad.NonreducedFraction((3, 6))
        NonreducedFraction(3, 6)

    ..  container:: example

        Initializes with an integer singleton:

        >>> abjad.NonreducedFraction((3,))
        NonreducedFraction(3, 1)

    ..  container:: example

        Nonreduced fractions inherit from built-in fraction:

        >>> isinstance(abjad.NonreducedFraction(3, 6), abjad.Fraction)
        True

    ..  container:: example

        Nonreduced fractions are numbers:

        >>> import numbers

        >>> isinstance(abjad.NonreducedFraction(3, 6), numbers.Number)
        True

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_numerator',
        '_denominator',
        )

    ### CONSTRUCTOR ###

    def __new__(class_, *arguments):
        import abjad
        is_fraction_like = False
        if len(arguments) == 1:
            try:
                numerator = arguments[0].numerator
                denominator = arguments[0].denominator
                is_fraction_like = True
            except AttributeError:
                pass
        if is_fraction_like:
            pass
        elif len(arguments) == 1 and isinstance(arguments[0], int):
            numerator = arguments[0]
            denominator = 1
        elif (len(arguments) == 1 and
            isinstance(arguments[0], tuple) and
            len(arguments[0]) == 1):
            numerator = arguments[0][0]
            denominator = 1
        elif (len(arguments) == 1 and
            isinstance(arguments[0], tuple) and
            isinstance(arguments[0][0], int) and
            isinstance(arguments[0][1], int)):
            numerator, denominator = arguments[0]
        elif len(arguments) == 1 and isinstance(arguments[0], str):
            numerator, denominator = class_._parse_input_string(arguments[0])
        elif (isinstance(arguments, tuple) and
            len(arguments) == 2 and
            isinstance(arguments[0], int) and
            isinstance(arguments[1], int)):
            numerator = arguments[0]
            denominator = arguments[1]
        elif len(arguments) == 0:
            numerator = 0
            denominator = 1
        else:
            message = 'can not initialize {}: {!r}.'
            message = message.format(class_.__name__, arguments)
            raise ValueError(message)
        numerator *= abjad.mathtools.sign(denominator)
        denominator = abs(denominator)
        self = fractions.Fraction.__new__(class_, numerator, denominator)
        self._numerator = numerator
        self._denominator = denominator
        return self

    def __init__(self, *arguments):
        """
        Dummy initializer to satisfy mypy.
        """
        pass

    ### SPECIAL METHODS ###

    def __abs__(self):
        """
        Gets absolute value of nonreduced fraction.

        ..  container:: example

            >>> abs(abjad.NonreducedFraction(-3, 3))
            NonreducedFraction(3, 3)

        Returns nonreduced fraction.
        """
        pair = (abs(self.numerator), self.denominator)
        return self._from_pair(pair)

    def __add__(self, argument):
        """
        Adds ``argument`` to nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) + 1
            NonreducedFraction(6, 3)

            >>> 1 + abjad.NonreducedFraction(3, 3)
            NonreducedFraction(6, 3)

        Returns nonreduced fraction.
        """
        import abjad
        if isinstance(argument, int):
            numerator = self.numerator + argument * self.denominator
            pair = (numerator, self.denominator)
            return self._from_pair(pair)
        if getattr(argument, 'denominator', 'foo') == 'foo':
            raise ValueError(argument)
        if self.denominator == argument.denominator:
            numerator = self.numerator + argument.numerator
            pair = (numerator, self.denominator)
            return self._from_pair(pair)
        else:
            denominators = [self.denominator, argument.denominator]
            denominator = abjad.mathtools.least_common_multiple(*denominators)
            self_multiplier = denominator // self.denominator
            argument_multiplier = denominator // argument.denominator
            self_numerator = self_multiplier * self.numerator
            argument_numerator = argument_multiplier * argument.numerator
            pair = (self_numerator + argument_numerator, denominator)
            return self._from_pair(pair)

    def __div__(self, argument):
        """
        Divides nonreduced fraction by ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) / 1
            NonreducedFraction(3, 3)

        Returns nonreduced fraction.
        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() / argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __eq__(self, argument):
        """
        Is true when ``argument`` equals nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) == 1
            True

        Returns true or false.
        """
        return self.reduce() == argument

    def __format__(self, format_specification=''):
        """
        Formats nonreduced fraction.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(-6, 3)
            >>> print(format(fraction))
            abjad.NonreducedFraction(-6, 3)

        Returns string.
        """
        if format_specification in ('', 'storage'):
            return system.StorageFormatManager(self).get_storage_format()
        return str(self)

    def __ge__(self, argument):
        """
        Is true when nonreduced fraction is greater than or equal to
        ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) >= 1
            True

        Returns true or false.
        """
        return self.reduce() >= argument

    def __gt__(self, argument):
        """
        Is true when nonreduced fraction is greater than ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) > 1
            False

        Returns true or false.
        """
        return self.reduce() > argument

    def __hash__(self):
        """
        Hashes nonreduced fraction.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        """
        return super().__hash__()

    def __le__(self, argument):
        """
        Is true when nonreduced fraction is less than or equal to ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) <= 1
            True

        Returns true or false.
        """
        return self.reduce() <= argument

    def __lt__(self, argument):
        """
        Is true when nonreduced fraction is less than ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) < 1
            False

        Returns true or false.
        """
        return self.reduce() < argument

    def __mul__(self, argument):
        """
        Multiplies nonreduced fraction by ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) * 3
            NonreducedFraction(9, 3)

        Returns nonreduced fraction.
        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() * argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __neg__(self):
        """
        Negates nonreduced fraction.

        ..  container:: example

            >>> -abjad.NonreducedFraction(3, 3)
            NonreducedFraction(-3, 3)

        Returns nonreduced fraction.
        """
        pair = (-self.numerator, self.denominator)
        return self._from_pair(pair)

    def __pow__(self, argument):
        """
        Raises nonreduced fraction to ``argument``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6) ** -1
            NonreducedFraction(6, 3)

        Returns nonreduced fraction.
        """
        if argument == -1:
            pair = (self.denominator, self.numerator)
            return self._from_pair(pair)
        return super().__pow__(argument)

    def __radd__(self, argument):
        """
        Adds nonreduced fraction to ``argument``.

        ..  container:: example

            >>> 1 + abjad.NonreducedFraction(3, 3)
            NonreducedFraction(6, 3)

        Returns nonreduced fraction.
        """
        return self + argument

    def __rdiv__(self, argument):
        """
        Divides ``argument`` by nonreduced fraction.

        ..  container:: example

            >>> 1 / abjad.NonreducedFraction(3, 3)
            NonreducedFraction(3, 3)

        Returns nonreduced fraction.
        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = argument / self.reduce()
        return self._fraction_with_denominator(fraction, max(denominators))

    __rtruediv__ = __rdiv__

    def __repr__(self):
        """
        Gets interpreter representation of nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6)
            NonreducedFraction(3, 6)

        Returns string.
        """
        return system.StorageFormatManager(self).get_repr_format()

    def __rmul__(self, argument):
        """
        Multiplies ``argument`` by nonreduced fraction.

        ..  container:: example

            >>> 3 * abjad.NonreducedFraction(3, 3)
            NonreducedFraction(9, 3)

        Returns nonreduced fraction.
        """
        return self * argument

    def __rsub__(self, argument):
        """
        Subtracts nonreduced fraction from ``argument``.

        ..  container:: example

            >>> 1 - abjad.NonreducedFraction(3, 3)
            NonreducedFraction(0, 3)

        Returns nonreduced fraction.
        """
        return -self + argument

    def __str__(self):
        """
        Gets string representation of nonreduced fraction.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(-6, 3)

            >>> str(fraction)
            '-6/3'

        Returns string.
        """
        return '{}/{}'.format(self.numerator, self.denominator)

    def __sub__(self, argument):
        """
        Subtracts ``argument`` from nonreduced fraction.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 3) - 2
            NonreducedFraction(-3, 3)

            >>> abjad.NonreducedFraction(5, 4) - abjad.NonreducedFraction(18, 16)
            NonreducedFraction(2, 16)

            >>> abjad.NonreducedFraction(18, 16) - abjad.NonreducedFraction(5, 4)
            NonreducedFraction(-2, 16)

        Returns nonreduced fraction.
        """
        denominators = [self.denominator]
        if isinstance(argument, type(self)):
            denominators.append(argument.denominator)
            argument = argument.reduce()
        fraction = self.reduce() - argument
        return self._fraction_with_denominator(fraction, max(denominators))

    def __truediv__(self, argument):
        """
        Divides nonreduced fraction in Python 3.

        Returns nonreduced fraction.
        """
        return self.__div__(argument)

    ### PRIVATE METHODS ###

    def _fraction_with_denominator(self, fraction, denominator):
        import abjad
        denominators = [denominator, fraction.denominator]
        denominator = abjad.mathtools.least_common_multiple(*denominators)
        result = self._from_pair(fraction)
        result = result.with_denominator(denominator)
        return result

    def _from_pair(self, pair):
        """
        Method is designed to be subclassed.
        """
        return type(self)(pair)

    def _get_format_specification(self):
        return system.FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_args_values=[
                self.numerator,
                self.denominator,
                ],
            storage_format_is_indented=False,
            storage_format_kwargs_names=[],
            )

    @staticmethod
    def _parse_input_string(string):
        if '/' in string:
            numerator, denominator = string.split('/')
            numerator = int(numerator)
            denominator = int(denominator)
        else:
            numerator = int(string)
            denominator = 1
        return numerator, denominator

    ### PUBLIC METHODS ###

    def multiply(self, multiplier, preserve_numerator=False):
        """
        Multiplies nonreduced fraction by ``multiplier`` with numerator
        preservation where possible.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(9, 16)

            >>> fraction.multiply((2, 3), preserve_numerator=True)
            NonreducedFraction(9, 24)

            >>> fraction.multiply((1, 2), preserve_numerator=True)
            NonreducedFraction(9, 32)

            >>> fraction.multiply((5, 6), preserve_numerator=True)
            NonreducedFraction(45, 96)

            >>> fraction = abjad.NonreducedFraction(3, 8)

            >>> fraction.multiply((2, 3), preserve_numerator=True)
            NonreducedFraction(3, 12)

        Returns nonreduced fraction.
        """
        import abjad
        if preserve_numerator:
            multiplier = abjad.Multiplier(multiplier)
            self_denominator = self.denominator
            candidate_result_denominator = self_denominator / multiplier
            if candidate_result_denominator.denominator == 1:
                pair = (self.numerator, candidate_result_denominator.numerator)
                return self._from_pair(pair)
            else:
                denominator = candidate_result_denominator.denominator
                result_numerator = self.numerator * denominator
                result_denominator = candidate_result_denominator.numerator
                pair = (result_numerator, result_denominator)
                return self._from_pair(pair)
        else:
            return multiplier * self

    def multiply_with_cross_cancelation(self, multiplier):
        """
        Multiplies nonreduced fraction by ``argument`` with cross-cancelation.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(4, 8)

            >>> fraction.multiply_with_cross_cancelation((2, 3))
            NonreducedFraction(4, 12)

            >>> fraction.multiply_with_cross_cancelation((4, 1))
            NonreducedFraction(4, 2)

            >>> fraction.multiply_with_cross_cancelation((3, 5))
            NonreducedFraction(12, 40)

            >>> fraction.multiply_with_cross_cancelation((6, 5))
            NonreducedFraction(12, 20)

            >>> fraction = abjad.NonreducedFraction(5, 6)
            >>> fraction.multiply_with_cross_cancelation((6, 5))
            NonreducedFraction(1, 1)

        Returns nonreduced fraction.
        """
        import abjad
        multiplier = abjad.Multiplier(multiplier)
        self_numerator_factors = abjad.mathtools.factors(self.numerator)
        multiplier_denominator_factors = abjad.mathtools.factors(
            multiplier.denominator)
        for factor in multiplier_denominator_factors[:]:
            if factor in self_numerator_factors:
                self_numerator_factors.remove(factor)
                multiplier_denominator_factors.remove(factor)
        self_denominator_factors = abjad.mathtools.factors(self.denominator)
        multiplier_numerator_factors = abjad.mathtools.factors(
            multiplier.numerator)
        for factor in multiplier_numerator_factors[:]:
            if factor in self_denominator_factors:
                self_denominator_factors.remove(factor)
                multiplier_numerator_factors.remove(factor)
        result_numerator_factors = self_numerator_factors + \
            multiplier_numerator_factors
        result_denominator_factors = self_denominator_factors + \
            multiplier_denominator_factors
        result_numerator = 1
        for factor in result_numerator_factors:
            result_numerator *= factor
        result_denominator = 1
        for factor in result_denominator_factors:
            result_denominator *= factor
        pair = (result_numerator, result_denominator)
        return self._from_pair(pair)

    def multiply_without_reducing(self, argument):
        """
        Multiplies nonreduced fraction by ``argument`` without reducing.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(3, 8)

            >>> fraction.multiply_without_reducing((3, 3))
            NonreducedFraction(9, 24)

            >>> fraction = abjad.NonreducedFraction(4, 8)

            >>> fraction.multiply_without_reducing((4, 5))
            NonreducedFraction(16, 40)

            >>> fraction.multiply_without_reducing((3, 4))
            NonreducedFraction(12, 32)

        Returns nonreduced fraction.
        """
        argument = self._from_pair(argument)
        numerator = self.numerator * argument.numerator
        denominator = self.denominator * argument.denominator
        pair = (numerator, denominator)
        return self._from_pair(pair)

    def reduce(self):
        """
        Reduces nonreduced fraction.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(-6, 3)

            >>> fraction.reduce()
            Fraction(-2, 1)

        Returns fraction.
        """
        return fractions.Fraction(self.numerator, self.denominator)

    def with_denominator(self, denominator):
        """
        Returns new nonreduced fraction with integer ``denominator``.

        ..  container:: example

            >>> abjad.NonreducedFraction(3, 6).with_denominator(12)
            NonreducedFraction(6, 12)

        ..  container:: example

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(12))
            ...
            0/6 0/12
            1/6 2/12
            2/6 4/12
            3/6 6/12
            4/6 8/12
            5/6 10/12
            6/6 12/12
            7/6 14/12
            8/6 16/12
            9/6 18/12
            10/6 20/12
            11/6 22/12

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(8))
            ...
            0/6 0/8
            1/6 1/6
            2/6 2/6
            3/6 4/8
            4/6 4/6
            5/6 5/6
            6/6 8/8
            7/6 7/6
            8/6 8/6
            9/6 12/8
            10/6 10/6
            11/6 11/6

            >>> for numerator in range(12):
            ...     fraction = abjad.NonreducedFraction(numerator, 6)
            ...     print(fraction, fraction.with_denominator(12))
            ...
            0/6 0/12
            1/6 2/12
            2/6 4/12
            3/6 6/12
            4/6 8/12
            5/6 10/12
            6/6 12/12
            7/6 14/12
            8/6 16/12
            9/6 18/12
            10/6 20/12
            11/6 22/12

        Returns nonreduced fraction.
        """
        import abjad
        current_numerator, current_denominator = self.pair
        multiplier = abjad.Multiplier(denominator, current_denominator)
        new_numerator = multiplier * current_numerator
        new_denominator = multiplier * current_denominator
        if (new_numerator.denominator == 1 and
            new_denominator.denominator == 1):
            pair = (new_numerator.numerator, new_denominator.numerator)
        else:
            pair = (current_numerator, current_denominator)
        return self._from_pair(pair)

    def with_multiple_of_denominator(self, denominator):
        """
        Returns new nonreduced fraction with multiple of integer
        ``denominator``.

        ..  container:: example

            >>> fraction = abjad.NonreducedFraction(3, 6)

            >>> fraction.with_multiple_of_denominator(5)
            NonreducedFraction(5, 10)

        ..  container:: example

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(2)
            NonreducedFraction(1, 2)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(4)
            NonreducedFraction(2, 4)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(8)
            NonreducedFraction(4, 8)

            >>> abjad.NonreducedFraction(1, 2).with_multiple_of_denominator(16)
            NonreducedFraction(8, 16)

        Returns nonreduced fraction.
        """
        result = self.with_denominator(denominator)
        while not result.denominator == denominator:
            denominator *= 2
            result = result.with_denominator(denominator)
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def denominator(self):
        """
        Denominator of nonreduced fraction.

        >>> fraction = abjad.NonreducedFraction(-6, 3)

        >>> fraction.denominator
        3

        Returns positive integer.
        """
        return self._denominator

    @property
    def imag(self):
        """
        Nonreduced fractions have no imaginary part.

        >>> fraction = abjad.NonreducedFraction(-6, 3)

        >>> fraction.imag
        0

        Returns zero.
        """
        return 0

    @property
    def numerator(self):
        """
        Numerator of nonreduced fraction.

        >>> fraction = abjad.NonreducedFraction(-6, 3)

        >>> fraction.numerator
        -6

        Returns integer.
        """
        return self._numerator

    @property
    def pair(self):
        """
        Read only pair of nonreduced fraction numerator and denominator.

        >>> fraction = abjad.NonreducedFraction(-6, 3)

        >>> fraction.pair
        (-6, 3)

        Returns integer pair.
        """
        return self.numerator, self.denominator

    @property
    def real(self):
        """
        Nonreduced fractions are their own real component.

        >>> fraction = abjad.NonreducedFraction(-6, 3)

        >>> fraction.real
        NonreducedFraction(-6, 3)

        Returns nonreduced fraction.
        """
        return self



from .ContextManager import ContextManager


class ForbidUpdate(ContextManager):
    r"""
    A context manager for forbidding score updates.

    ..  container:: example

        >>> staff = abjad.Staff("c'8 d'8 ~ d'2 e'4")
        >>> with abjad.ForbidUpdate(component=staff):
        ...     for note in staff[:]:
        ...         pitch_1 = note.written_pitch
        ...         pitch_2 = pitch_1 + abjad.NamedInterval('M3')
        ...         pitches = [pitch_1, pitch_2]
        ...         chord = abjad.Chord(pitches, note.written_duration)
        ...         abjad.mutate(note).replace(chord)
        ...

        >>> abjad.inspect(staff).is_well_formed()
        True

        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff {
                <c' e'>8
                <d' fs'>8 ~
                <d' fs'>2
                <e' gs'>4
            }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Context managers'

    __slots__ = (
        '_component',
        '_update_on_enter',
        '_update_on_exit',
        )

    ### INITIALIZER ###

    def __init__(
        self,
        component=None,
        update_on_enter=True,
        update_on_exit=None,
        ):
        import abjad
        prototype = (abjad.Component, type(None))
        assert isinstance(component, prototype)
        self._component = component
        if update_on_enter is not None:
            update_on_enter = bool(update_on_enter)
        self._update_on_enter = update_on_enter
        if update_on_exit is not None:
            update_on_exit = bool(update_on_exit)
        self._update_on_exit = update_on_exit

    ### SPECIAL METHODS ###

    def __enter__(self):
        """
        Enters context manager.

        Returns context manager.
        """
        if self.component is not None:
            self.component._update_now(offsets=True)
            self.component._is_forbidden_to_update = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exits context manager.

        Returns none.
        """
        if self.component is not None:
            self.component._is_forbidden_to_update = False
            if self.update_on_exit:
                self.component._update_now(offsets=True)

    ### PUBLIC PROPERTIES ###

    @property
    def component(self):
        """
        Gets component.

        Set to component or none.

        Returns component or none.
        """
        return self._component

    @property
    def update_on_enter(self):
        """
        Is true when context manager should update offsets on enter.

        Set to true, false or none.

        Returns true, false or none.
        """
        return self._update_on_enter

    @property
    def update_on_exit(self):
        """
        Is true when context manager should update offsets on exit.

        Set to true, false or none.

        Returns true, false or none.
        """
        return self._update_on_exit



def list_all_ide_functions(modules=None):
    r'''Lists all public functions defined in Abjad IDE.

    ::

        >>> all_functions = utilities.list_all_ide_functions()  # doctest: +SKIP

    '''
    from abjad import utilities
    try:
        return utilities.list_all_functions(modules='ide')
    except ImportError:
        return []



import numbers
from .NamedIntervalClass import NamedIntervalClass


class NamedInversionEquivalentIntervalClass(NamedIntervalClass):
    """
    Named inversion-equivalent interval-class.

    ..  container:: example

        Initializes from string:

        >>> abjad.NamedInversionEquivalentIntervalClass('-m14')
        NamedInversionEquivalentIntervalClass('+M2')

    ..  container:: example

        Initializes from pair:

        >>> abjad.NamedInversionEquivalentIntervalClass(('perfect', 1))
        NamedInversionEquivalentIntervalClass('P1')

        >>> abjad.NamedInversionEquivalentIntervalClass(('perfect', -1))
        NamedInversionEquivalentIntervalClass('P1')

        >>> abjad.NamedInversionEquivalentIntervalClass(('augmented', 4))
        NamedInversionEquivalentIntervalClass('+A4')

        >>> abjad.NamedInversionEquivalentIntervalClass(('augmented', -4))
        NamedInversionEquivalentIntervalClass('+A4')

        >>> abjad.NamedInversionEquivalentIntervalClass(('augmented', 11))
        NamedInversionEquivalentIntervalClass('+A4')

        >>> abjad.NamedInversionEquivalentIntervalClass(('augmented', -11))
        NamedInversionEquivalentIntervalClass('+A4')

    ..  container:: example

        Initializes from other interval-class:

        >>> interval_class = abjad.NamedInversionEquivalentIntervalClass(
        ...     'P1',
        ...     )
        >>> abjad.NamedInversionEquivalentIntervalClass(interval_class)
        NamedInversionEquivalentIntervalClass('P1')

    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, name='P1'):
        super().__init__(name or 'P1')
        self._quality, self._number = self._process_quality_and_number(
            self._quality,
            self._number,
            )

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when `argument` is a named inversion-equivalent
        interval-class with name equal to that of this named
        inversion-equivalent interval-class.

        ..  container:: example

            >>> class_ = abjad.NamedInversionEquivalentIntervalClass
            >>> interval_class_1 = class_('P1')
            >>> interval_class_2 = class_('P1')
            >>> interval_class_3 = class_('m2')

            >>> interval_class_1 == interval_class_1
            True
            >>> interval_class_1 == interval_class_2
            True
            >>> interval_class_1 == interval_class_3
            False

            >>> interval_class_2 == interval_class_1
            True
            >>> interval_class_2 == interval_class_2
            True
            >>> interval_class_2 == interval_class_3
            False

            >>> interval_class_3 == interval_class_1
            False
            >>> interval_class_3 == interval_class_2
            False
            >>> interval_class_3 == interval_class_3
            True

        Returns true or false.
        """
        return super().__eq__( argument)

    def __hash__(self):
        """
        Hashes named inversion-equivalent interval-class.

        Returns integer.
        """
        return super().__hash__()

    ### PRIVATE METHODS ###

    @classmethod
    def _invert_quality_string(class_, quality):
        inversions = {'M': 'm', 'm': 'M', 'P': 'P'}
        if quality in inversions:
            return inversions[quality]
        if quality[0] == 'A':
            return 'd' * len(quality)
        return 'A' * len(quality)

    @classmethod
    def _is_representative_number(class_, argument):
        if isinstance(argument, numbers.Number):
            if 1 <= argument <= 4 or argument == 8:
                return True
        return False

    @classmethod
    def _process_quality_and_number(class_, quality, number):
        if number == 0:
            message = 'named interval can not equal zero.'
            raise ValueError(message)
        elif abs(number) == 1:
            number = 1
        elif abs(number) % 7 == 0:
            number = 7
        elif abs(number) % 7 == 1:
            number = 8
        else:
            number = abs(number) % 7
        if class_._is_representative_number(number):
            quality = quality
            number = number
        else:
            quality = class_._invert_quality_string(quality)
            number = 9 - number
        return quality, number

    ### PUBLIC METHODS ###

    @classmethod
    def from_pitch_carriers(class_, pitch_carrier_1, pitch_carrier_2):
        """
        Makes named inversion-equivalent interval-class from
        `pitch_carrier_1` and `pitch_carrier_2`.

        ..  container:: example

            >>> class_ = abjad.NamedInversionEquivalentIntervalClass
            >>> class_.from_pitch_carriers(
            ...     abjad.NamedPitch(-2),
            ...     abjad.NamedPitch(12),
            ...     )
            NamedInversionEquivalentIntervalClass('+M2')

        Returns new named inversion-equivalent interval-class.
        """
        import abjad
        named_interval = abjad.NamedInterval.from_pitch_carriers(
            pitch_carrier_1,
            pitch_carrier_2,
            )
        string = str(named_interval)
        return class_(string)



from .Block import Block


class ContextBlock(Block):
    r"""
    A LilyPond file ``\context`` block.

    ..  container:: example

        >>> block = abjad.ContextBlock(
        ...     source_lilypond_type='Staff',
        ...     name='FluteStaff',
        ...     type_='Engraver_group',
        ...     alias='Staff',
        ...     )
        >>> block.remove_commands.append('Forbid_line_break_engraver')
        >>> block.consists_commands.append('Horizontal_bracket_engraver')
        >>> block.accepts_commands.append('FluteUpperVoice')
        >>> block.accepts_commands.append('FluteLowerVoice')
        >>> block.items.append(r'\accidentalStyle dodecaphonic')
        >>> abjad.override(block).beam.positions = (-4, -4)
        >>> abjad.override(block).stem.stem_end_position = -6
        >>> abjad.setting(block).auto_beaming = False
        >>> abjad.setting(block).tuplet_full_length = True
        >>> block
        <ContextBlock(source_lilypond_type='Staff', name='FluteStaff', type_='Engraver_group', alias='Staff')>

        >>> print(format(block))
        \context {
            \Staff
            \name FluteStaff
            \type Engraver_group
            \alias Staff
            \remove Forbid_line_break_engraver
            \consists Horizontal_bracket_engraver
            \accepts FluteUpperVoice
            \accepts FluteLowerVoice
            \override Beam.positions = #'(-4 . -4)
            \override Stem.stem-end-position = #-6
            autoBeaming = ##f
            tupletFullLength = ##t
            \accidentalStyle dodecaphonic
        }

    """

    ### INITIALIZER ###

    def __init__(
        self,
        source_lilypond_type=None,
        name=None,
        type_=None,
        alias=None,
        ):
        Block.__init__(self, name='context')
        self._source_lilypond_type = source_lilypond_type
        self._name = name
        self._type_ = type_
        self._alias = alias
        self._accepts_commands = []
        self._consists_commands = []
        self._remove_commands = []

    ### PRIVATE METHODS ###

    def _get_format_pieces(self):
        import abjad
        indent = abjad.LilyPondFormatManager.indent
        result = []
        string = '{} {{'.format(self._escaped_name)
        result.append(string)
        manager = abjad.LilyPondFormatManager
        # CAUTION: source context name must come before type_ to allow
        # context redefinition.
        if self.source_lilypond_type is not None:
            string = indent + r'\{}'.format(self.source_lilypond_type)
            result.append(string)
        if self.name is not None:
            string = indent + r'\name {}'.format(self.name)
            result.append(string)
        if self.type_ is not None:
            string = indent + r'\type {}'.format(self.type_)
            result.append(string)
        if self.alias is not None:
            string = indent + r'\alias {}'.format(self.alias)
            result.append(string)
        for statement in self.remove_commands:
            string = indent + r'\remove {}'.format(statement)
            result.append(string)
        # CAUTION: LilyPond \consists statements are order-significant!
        for statement in self.consists_commands:
            string = indent + r'\consists {}'.format(statement)
            result.append(string)
        for statement in self.accepts_commands:
            string = indent + r'\accepts {}'.format(statement)
            result.append(string)
        overrides = abjad.override(self)._list_format_contributions('override')
        for statement in overrides:
            string = indent + statement
            result.append(string)
        setting_contributions = []
        for key, value in abjad.setting(self)._get_attribute_tuples():
            setting_contribution = \
                manager.format_lilypond_context_setting_in_with_block(
                    key, value)
            setting_contributions.append(setting_contribution)
        for setting_contribution in sorted(setting_contributions):
            string = indent + setting_contribution
            result.append(string)
        for item in self.items:
            if isinstance(item, str):
                string = indent + '{}'.format(item)
                result.append(string)
            elif '_get_format_pieces' in dir(item):
                pieces = item._get_format_pieces()
                pieces = [indent + item for item in pieces]
                result.extend(pieces)
            else:
                pass
        result.append('}')
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def accepts_commands(self):
        r"""
        Gets arguments of LilyPond ``\accepts`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.accepts_commands
            ['FluteUpperVoice', 'FluteLowerVoice']

        Returns list.
        """
        return self._accepts_commands

    @property
    def alias(self):
        r"""
        Gets and sets argument of LilyPond ``\alias`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.alias
            'Staff'

        Returns string or none.
        """
        return self._alias

    @property
    def consists_commands(self):
        r"""
        Gets arguments of LilyPond ``\consists`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.consists_commands
            ['Horizontal_bracket_engraver']

        Returns list.
        """
        return self._consists_commands

    @property
    def items(self):
        r"""
        Gets items in context block.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.items
            ['\\accidentalStyle dodecaphonic']

        Returns list.
        """
        return self._items

    @property
    def name(self):
        r"""
        Gets and sets argument of LilyPond ``\name`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.name
            'FluteStaff'

        Returns string or none.
        """
        return self._name

    @property
    def remove_commands(self):
        r"""
        Gets arguments of LilyPond ``\remove`` commands.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.remove_commands
            ['Forbid_line_break_engraver']

        Returns list.
        """
        return self._remove_commands

    @property
    def source_lilypond_type(self):
        r"""
        Gets and sets source context name.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.source_lilypond_type
            'Staff'

        Returns string or none.
        """
        return self._source_lilypond_type

    @property
    def type_(self):
        r"""
        Gets and sets argument of LilyPond ``\type`` command.

        ..  container:: example

            >>> block = abjad.ContextBlock(
            ...     source_lilypond_type='Staff',
            ...     name='FluteStaff',
            ...     type_='Engraver_group',
            ...     alias='Staff',
            ...     )
            >>> block.remove_commands.append('Forbid_line_break_engraver')
            >>> block.consists_commands.append('Horizontal_bracket_engraver')
            >>> block.accepts_commands.append('FluteUpperVoice')
            >>> block.accepts_commands.append('FluteLowerVoice')
            >>> block.items.append(r'\accidentalStyle dodecaphonic')
            >>> abjad.override(block).beam.positions = (-4, -4)
            >>> abjad.override(block).stem.stem_end_position = -6
            >>> abjad.setting(block).auto_beaming = False
            >>> abjad.setting(block).tuplet_full_length = True

            >>> block.type_
            'Engraver_group'

        Returns string or none.
        """
        return self._type_
